/* Copyright (C) 2003 MySQL AB
   Copyright (C) 2008 DDengine.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define MYSQL_SERVER 1
#include "mysql_priv.h"
#include "ha_revision.h"
#include <mysql/plugin.h>

#include "single_table_handler.h"
#include "double_table_handler.h"

static handler *revision_create_handler(handlerton *hton,
                                       TABLE_SHARE *table, 
                                       MEM_ROOT *mem_root);

handlerton *revision_hton;

/* Variables for revision share methods */
static HASH revision_open_tables; ///< Hash used to track the number of open tables; variable for revision share methods
pthread_mutex_t revision_mutex;   ///< This is the mutex used to init the hash; variable for revision share methods
static int revision_init= 0;      ///< This variable is used to check the init state of hash; variable for revision share methods

/** @brief
  Function we use in the creation of our hash to get key.
*/
static uchar* revision_get_key(REVISION_SHARE *share,uint *length,
                             my_bool not_used __attribute__((unused)))
{
  *length=share->table_name_length;
  return (uchar*) share->table_name;
}


static int revision_init_func(void *p)
{
  DBUG_ENTER("revision_init_func");

  revision_hton= (handlerton *)p;
  VOID(pthread_mutex_init(&revision_mutex,MY_MUTEX_INIT_FAST));
  (void) hash_init(&revision_open_tables,system_charset_info,32,0,0,
                   (hash_get_key) revision_get_key,0,0);

  revision_hton->state=   SHOW_OPTION_YES;
  revision_hton->db_type= DB_TYPE_UNKNOWN;
  revision_hton->create=  revision_create_handler;
  revision_hton->flags=   HTON_CAN_RECREATE;

  DBUG_RETURN(0);
}


static int revision_done_func(void *)
{
  int error= 0;
  DBUG_ENTER("revision_done_func");

  if (revision_open_tables.records)
    error= 1;
  hash_free(&revision_open_tables);
  pthread_mutex_destroy(&revision_mutex);

  DBUG_RETURN(0);
}


/** @brief
  Revision of simple lock controls. The "share" it creates is a structure we
  will pass to each revision handler. Do you have to have one of these? Well,
  you have pieces that are used for locking, and they are needed to function.
*/
static REVISION_SHARE *get_share(const char *table_name, TABLE *)
{
  REVISION_SHARE *share;
  uint length;
  char *tmp_name;

  pthread_mutex_lock(&revision_mutex);
  length=(uint) strlen(table_name);

  if (!(share=(REVISION_SHARE*) hash_search(&revision_open_tables,
                                           (uchar*) table_name,
                                           length)))
  {
    if (!(share=(REVISION_SHARE *)
          my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
                          &share, sizeof(*share),
                          &tmp_name, length+1,
                          NullS)))
    {
      pthread_mutex_unlock(&revision_mutex);
      return NULL;
    }

    share->use_count=0;
    share->table_name_length=length;
    share->table_name=tmp_name;
    strmov(share->table_name,table_name);
    if (my_hash_insert(&revision_open_tables, (uchar*) share))
      goto error;
    thr_lock_init(&share->lock);
    pthread_mutex_init(&share->mutex,MY_MUTEX_INIT_FAST);
  }
  share->use_count++;
  pthread_mutex_unlock(&revision_mutex);

  return share;

error:
  pthread_mutex_destroy(&share->mutex);
  my_free(share, MYF(0));

  return NULL;
}


/** @brief
  Free lock controls. We call this whenever we close a table. If the table had
  the last reference to the share, then we free memory associated with it.
*/
static int free_share(REVISION_SHARE *share)
{
  pthread_mutex_lock(&revision_mutex);
  if (!--share->use_count)
  {
    hash_delete(&revision_open_tables, (uchar*) share);
    thr_lock_delete(&share->lock);
    pthread_mutex_destroy(&share->mutex);
    my_free(share, MYF(0));
  }
  pthread_mutex_unlock(&revision_mutex);

  return 0;
}

static handler* revision_create_handler(handlerton *hton,
                                       TABLE_SHARE *table, 
                                       MEM_ROOT *mem_root)
{
  return new (mem_root) ha_revision(hton, table, mem_root);
}


/** Revision handler factory.
    So far only single table handlers is implemented.
*/
static revision_handler *instantiate_revision_handler(TABLE_SHARE *table_arg,
                                                      MEM_ROOT *mem_root)
{
  /* Uses the comment field to determine which handler is to be instantiated
  */
  if (table_arg->comment.length != 0)
  {
    if (strstr(table_arg->comment.str, "DOUBLE") != NULL)
    {
      return new double_table_handler(table_arg, mem_root);
    }
  }
  return new single_table_handler(table_arg, mem_root);
}


static int check_select_mode(MYSQL_THD thd,
                             struct st_mysql_sys_var *var,
                             void *save, struct st_mysql_value *value);


static MYSQL_THDVAR_STR(select_mode,
                        PLUGIN_VAR_RQCMDARG
                        | PLUGIN_VAR_MEMALLOC,
                        "Specify the revision(s) to be selected.",
                        check_select_mode, NULL,
                        select_current);


ha_revision::ha_revision(handlerton *hton, TABLE_SHARE *table_arg,
                         MEM_ROOT *root)
  : handler(hton, table_arg), uhandler(NULL), mem_root (root)
{
  if (table_arg && table_arg->db_type())
  {
    uhandler= instantiate_revision_handler(table_arg, mem_root);
  }
}


ha_revision::~ha_revision()
{
  delete uhandler;
}


/** @brief
  If frm_error() is called then we will use this to determine the file
  extensions that exist for the storage engine. This is also used by the
  default rename_table and delete_table method in handler.cc.

    @see
  rename_table method in handler.cc and
  delete_table method in handler.cc
*/
static const char *ha_revision_exts[] = {
  ".REV",
  NullS
};


const char **ha_revision::bas_ext() const
{
  return ha_revision_exts;
}


/** @brief
    The name of the index type that will be used for display.
    Don't implement this method unless you really have indexes.
*/
const char *ha_revision::index_type(uint inx)
{
  return uhandler->get_uhandler()->index_type(inx);
}


/** @brief
    This is a list of flags that indicate what functionality the storage engine
    implements. The current table flags are documented in handler.h
*/
ulonglong ha_revision::table_flags() const
{
  if (uhandler != NULL)
  {
    uhandler->get_uhandler()->init();

    /* Sets the ref/dup_ref as soon as possible after table open.
    */
    const_cast<ha_revision *>(this)->ref= uhandler->get_uhandler()->ref;
    const_cast<ha_revision *>(this)->dup_ref=
      uhandler->get_uhandler()->dup_ref;
    const_cast<ha_revision *>(this)->ref_length=
      uhandler->get_uhandler()->ref_length;
    return uhandler->get_uhandler()->ha_table_flags();
  }

  /* Use comment field to create the right handler and get the underlying
     table flags.
  */
  THD *thd= current_thd;
  if (SQLCOM_CREATE_TABLE == thd->lex->sql_command)
  {
    handlerton *underlying= NULL;
    if (0 != thd->lex->create_info.comment.length)
    {
      plugin_ref plugin=
        ha_resolve_by_name(thd, &thd->lex->create_info.comment);
      if (plugin)
        underlying= plugin_data(plugin, handlerton *);
    }
    if (!underlying)
      underlying= ha_default_handlerton(thd);

    handler *h= get_new_handler(NULL, thd->mem_root,
                                underlying);
    ulonglong f= h->ha_table_flags();
    delete h;
    return(f);
  }

  return (HA_PRIMARY_KEY_IN_READ_INDEX | HA_FILE_BASED
          | HA_REC_NOT_IN_SEQ | HA_AUTO_PART_KEY | HA_CAN_INDEX_BLOBS
          | HA_PRIMARY_KEY_REQUIRED_FOR_DELETE
          | HA_PARTIAL_COLUMN_READ | HA_NULL_IN_KEY);
}


void ha_revision::start_bulk_insert(ha_rows rows)
{
  uhandler->get_uhandler()->ha_start_bulk_insert(rows);
}


int ha_revision::end_bulk_insert()
{
  return uhandler->get_uhandler()->ha_end_bulk_insert();
}


int ha_revision::disable_indexes(uint mode)
{
  return uhandler->get_uhandler()->ha_disable_indexes(mode);
}


int ha_revision::enable_indexes(uint mode)
{
  return uhandler->get_uhandler()->ha_enable_indexes(mode);
}


int ha_revision::indexes_are_disabled(void)
{
  return uhandler->get_uhandler()->indexes_are_disabled();
}


/** @brief
    This is a bitmap of flags that indicates how the storage engine
    implements indexes. The current index flags are documented in
    handler.h. If you do not implement indexes, just return zero here.

    @details
    part is the key part to check. First key part is 0.
    If all_parts is set, MySQL wants to know the flags for the combined
    index, up to and including 'part'.
*/
ulong ha_revision::index_flags(uint inx, uint part, bool all_parts) const
{
  if (uhandler != NULL)
    return uhandler->get_uhandler()->index_flags(inx, part, all_parts);
  return 0;
}


/** @brief
    unireg.cc will call max_supported_record_length(), max_supported_keys(),
    max_supported_key_parts(), uint max_supported_key_length()
    to make sure that the storage engine can handle the data it is about to
    send. Return *real* limits of your storage engine here; MySQL will do
    min(your_limits, MySQL_limits) automatically.
*/
uint ha_revision::max_supported_record_length() const
{
  if (uhandler != NULL)
    return uhandler->get_uhandler()->max_supported_record_length();
  return handler::max_supported_record_length();
}


/** @brief
    unireg.cc will call this to make sure that the storage engine can handle
    the data it is about to send. Return *real* limits of your storage engine
    here; MySQL will do min(your_limits, MySQL_limits) automatically.
    
    @details
    There is no need to implement ..._key_... methods if your engine doesn't
    support indexes.
*/
uint ha_revision::max_supported_keys() const
{
  if (uhandler != NULL)
    return uhandler->get_uhandler()->max_supported_keys();
  return MAX_KEY;
}


/** @brief
    unireg.cc will call this to make sure that the storage engine can handle
    the data it is about to send. Return *real* limits of your storage engine
    here; MySQL will do min(your_limits, MySQL_limits) automatically.

    @details
    There is no need to implement ..._key_... methods if your engine doesn't
    support indexes.
*/
uint ha_revision::max_supported_key_parts() const
{
  if (uhandler != NULL)
    return uhandler->get_uhandler()->max_supported_key_parts();
  return MAX_REF_PARTS;
}


/** @brief
    unireg.cc will call this to make sure that the storage engine can handle
    the data it is about to send. Return *real* limits of your storage engine
    here; MySQL will do min(your_limits, MySQL_limits) automatically.

    @details
    There is no need to implement ..._key_... methods if your engine doesn't
    support indexes.
*/
uint ha_revision::max_supported_key_length() const
{
  if (uhandler != NULL)
    return uhandler->get_uhandler()->max_supported_key_length();
  return MAX_KEY_LENGTH;
}


/** @brief
    Called in test_quick_select to determine if indexes should be used.
*/
double ha_revision::scan_time()
{
  return uhandler->scan_time();
}


/** @brief
    This method will never be called if you do not implement indexes.
*/
double ha_revision::read_time(uint index, uint ranges, ha_rows rows)
{
  return uhandler->read_time(index, ranges, rows);
}


const key_map *ha_revision::keys_to_use_for_scanning()
{
  return uhandler->get_uhandler()->keys_to_use_for_scanning();
}


/** @brief
  Used for opening tables. The name will be the name of the file.

    @details
  A table is opened when it needs to be opened; e.g. when a request comes in
  for a SELECT on the table (tables are not open and closed for each request,
  they are cached).

  Called from handler.cc by handler::ha_open(). The server opens all tables by
  calling ha_open() which then calls the handler specific open().

    @see
  handler::ha_open() in handler.cc
*/
int ha_revision::open(const char *name, int mode, uint test_if_locked)
{
  DBUG_ENTER("ha_revision::open");

  if (!(share = get_share(name, table)))
    DBUG_RETURN(1);
  thr_lock_data_init(&share->lock,&lock,NULL);

  DBUG_RETURN(uhandler->open(table, name, mode, test_if_locked));
}


/** @brief
  Closes a table. We call the free_share() function to free any resources
  that we have allocated in the "shared" structure.

    @details
  Called from sql_base.cc, sql_select.cc, and table.cc. In sql_select.cc it is
  only used to close up temporary tables or during the process where a
  temporary table is converted over to being a myisam table.

  For sql_base.cc look at close_data_tables().

    @see
  sql_base.cc, sql_select.cc and table.cc
*/
int ha_revision::close(void)
{
  DBUG_ENTER("ha_revision::close");
  uhandler->close();
  DBUG_RETURN(free_share(share));
}


bool ha_revision::get_error_message(int error, String *buf)
{
  return uhandler->get_error_message(error, buf);
}

/** @brief
  write_row() inserts a row. No extra() hint is given currently if a bulk load
  is happening. buf() is a uchar array of data. You can use the field
  information to extract the data from the native uchar array type.

    @details
  Example of this would be:
    @code
  for (Field **field=table->field ; *field ; field++)
  {
    ...
  }
    @endcode

  See ha_tina.cc for an example of extracting all of the data as strings.
  ha_berekly.cc has an example of how to store it intact by "packing" it
  for ha_berkeley's own native storage type.

  See the note for update_row() on auto_increments and timestamps. This
  case also applies to write_row().

  Called from item_sum.cc, item_sum.cc, sql_acl.cc, sql_insert.cc,
  sql_insert.cc, sql_select.cc, sql_table.cc, sql_udf.cc, and sql_update.cc.

    @see
  item_sum.cc, item_sum.cc, sql_acl.cc, sql_insert.cc,
  sql_insert.cc, sql_select.cc, sql_table.cc, sql_udf.cc and sql_update.cc
*/
int ha_revision::write_row(uchar * buf)
{
  return uhandler->write_row(buf);
}


/** @brief
  Yes, update_row() does what you expect, it updates a row. old_data will have
  the previous row record in it, while new_data will have the newest data in
  it.
  Keep in mind that the server can do updates based on ordering if an ORDER BY
  clause was used. Consecutive ordering is not guaranteed.

    @details
  Currently new_data will not have an updated auto_increament record, or
  and updated timestamp field. You can do these for example by doing:
    @code
  if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
    table->timestamp_field->set_time();
  if (table->next_number_field && record == table->record[0])
    update_auto_increment();
    @endcode

  Called from sql_select.cc, sql_acl.cc, sql_update.cc, and sql_insert.cc.

    @see
  sql_select.cc, sql_acl.cc, sql_update.cc and sql_insert.cc
*/
int ha_revision::update_row(const uchar * old_data, uchar * new_data)
{
  return uhandler->update_row(old_data, new_data);
}


/** @brief
  This will delete a row. buf will contain a copy of the row to be deleted.
  The server will call this right after the current row has been called (from
  either a previous rnd_nexT() or index call).

    @details
  If you keep a pointer to the last row or can access a primary key it will
  make doing the deletion quite a bit easier. Keep in mind that the server does
  not guarantee consecutive deletions. ORDER BY clauses can be used.

  Called in sql_acl.cc and sql_udf.cc to manage internal table information.
  Called in sql_delete.cc, sql_insert.cc, and sql_select.cc. In sql_select it
  is used for removing duplicates while in insert it is used for REPLACE calls.

    @see
  sql_acl.cc, sql_udf.cc, sql_delete.cc, sql_insert.cc and sql_select.cc
*/
int ha_revision::delete_row(const uchar * buf)
{
  DBUG_ENTER("ha_revision::delete_row");
  DBUG_RETURN(uhandler->delete_row(buf));
}


void ha_revision::set_select_mode()
{
  /* Do not set the select mode for update and delete
     statements as it changes the way index_read() works.
  */
  THD *thd= current_thd;
  if ((SQLCOM_UPDATE != thd->lex->sql_command)
      && (SQLCOM_DELETE != thd->lex->sql_command))
  {
    char *v= THDVAR(current_thd, select_mode);
    if (strcmp(v, select_current) == 0)
      uhandler->set_select_mode(revision_handler::current);
    else if (strcmp(v, select_history) == 0)
      uhandler->set_select_mode(revision_handler::history);
    else if (strcmp(v, select_deleted) == 0)
      uhandler->set_select_mode(revision_handler::deleted);
    else
    {
      int cut= 0;
      uint flags= 0;
      MYSQL_TIME time;
      enum_mysql_timestamp_type t=
        str_to_datetime(v, strlen(v), &time, flags, &cut);
      if ((MYSQL_TIMESTAMP_ERROR != t)
          && (MYSQL_TIMESTAMP_NONE != t))
      {
        uhandler->set_timestamp(time);
        uhandler->set_select_mode(revision_handler::timestamp);
      }
    }
  }
}


int ha_revision::index_init(uint idx, bool sorted)
{
  set_select_mode();
  active_index= idx;
  return uhandler->get_uhandler()->ha_index_init(idx, sorted);
}


int ha_revision::index_end()
{
  uhandler->set_select_mode(revision_handler::none);
  active_index= MAX_KEY;
  return uhandler->get_uhandler()->ha_index_end();
}


/** @brief
  Positions an index cursor to the index specified in the handle. Fetches the
  row if available. If the key value is null, begin at the first key of the
  index.
*/
int ha_revision::index_read(uchar * buf, const uchar * key,
                            uint key_len,
                            enum ha_rkey_function find_flag)
{
  DBUG_ENTER("ha_revision::index_read");
  DBUG_RETURN(uhandler->index_read(buf, key,
                                   key_len, find_flag));
}


/** @brief
  Used to read forward through the index.
*/
int ha_revision::index_next(uchar * buf)
{
  DBUG_ENTER("ha_revision::index_next");
  DBUG_RETURN(uhandler->index_next(buf));
}


/** @brief
  Used to read backwards through the index.
*/
int ha_revision::index_prev(uchar * buf)
{
  DBUG_ENTER("ha_revision::index_prev");
  DBUG_RETURN(uhandler->get_uhandler()->index_prev(buf));
}


/** @brief
  index_first() asks for the first key in the index.

    @details
  Called from opt_range.cc, opt_sum.cc, sql_handler.cc, and sql_select.cc.

    @see
  opt_range.cc, opt_sum.cc, sql_handler.cc and sql_select.cc
*/
int ha_revision::index_first(uchar * buf)
{
  DBUG_ENTER("ha_revision::index_first");
  DBUG_RETURN(uhandler->get_uhandler()->index_first(buf));
}


/** @brief
  index_last() asks for the last key in the index.

    @details
  Called from opt_range.cc, opt_sum.cc, sql_handler.cc, and sql_select.cc.

    @see
  opt_range.cc, opt_sum.cc, sql_handler.cc and sql_select.cc
*/
int ha_revision::index_last(uchar * buf)
{
  DBUG_ENTER("ha_revision::index_last");
  DBUG_RETURN(uhandler->get_uhandler()->index_last(buf));
}


/** @brief
  rnd_init() is called when the system wants the storage engine to do a table
  scan. See the example in the introduction at the top of this file to see when
  rnd_init() is called.

    @details
  Called from filesort.cc, records.cc, sql_handler.cc, sql_select.cc,
  sql_table.cc, and sql_update.cc.

    @see
  filesort.cc, records.cc, sql_handler.cc, sql_select.cc, sql_table.cc and
  sql_update.cc
*/
int ha_revision::rnd_init(bool scan)
{
  DBUG_ENTER("ha_revision::rnd_init");
  set_select_mode();
  DBUG_RETURN(uhandler->get_uhandler()->ha_rnd_init(scan));
}


int ha_revision::rnd_end()
{
  DBUG_ENTER("ha_revision::rnd_end");
  uhandler->set_select_mode(revision_handler::none);
  DBUG_RETURN(uhandler->get_uhandler()->ha_rnd_end());
}

/** @brief
  This is called for each row of the table scan. When you run out of records
  you should return HA_ERR_END_OF_FILE. Fill buff up with the row information.
  The Field structure for the table is the key to getting data into buf
  in a manner that will allow the server to understand it.

    @details
  Called from filesort.cc, records.cc, sql_handler.cc, sql_select.cc,
  sql_table.cc, and sql_update.cc.

    @see
  filesort.cc, records.cc, sql_handler.cc, sql_select.cc, sql_table.cc and
  sql_update.cc
*/
int ha_revision::rnd_next(uchar *buf)
{
  DBUG_ENTER("ha_revision::rnd_next");
  DBUG_RETURN(uhandler->rnd_next(buf));
}


/** @brief
  position() is called after each call to rnd_next() if the data needs
  to be ordered. You can do something like the following to store
  the position:
    @code
  my_store_ptr(ref, ref_length, current_position);
    @endcode

    @details
  The server uses ref to store data. ref_length in the above case is
  the size needed to store current_position. ref is just a uchar array
  that the server will maintain. If you are using offsets to mark rows, then
  current_position should be the offset. If it is a primary key like in
  BDB, then it needs to be a primary key.

  Called from filesort.cc, sql_select.cc, sql_delete.cc, and sql_update.cc.

    @see
  filesort.cc, sql_select.cc, sql_delete.cc and sql_update.cc
*/
void ha_revision::position(const uchar *record)
{
  DBUG_ENTER("ha_revision::position");
  uhandler->get_uhandler()->position(record);
  DBUG_VOID_RETURN;
}


/** @brief
  This is like rnd_next, but you are given a position to use
  to determine the row. The position will be of the type that you stored in
  ref. You can use ha_get_ptr(pos,ref_length) to retrieve whatever key
  or position you saved when position() was called.

    @details
  Called from filesort.cc, records.cc, sql_insert.cc, sql_select.cc, and
  sql_update.cc.

    @see
  filesort.cc, records.cc, sql_insert.cc, sql_select.cc and sql_update.cc
*/
int ha_revision::rnd_pos(uchar * buf, uchar *pos)
{
  DBUG_ENTER("ha_revision::rnd_pos");
  DBUG_RETURN(uhandler->get_uhandler()->rnd_pos(buf, pos));
}


/** @brief
  ::info() is used to return information to the optimizer. See my_base.h for
  the complete description.

    @details
  Currently this table handler doesn't implement most of the fields really
  needed.
  SHOW also makes use of this data.

  You will probably want to have the following in your code:
    @code
  if (records < 2)
    records = 2;
    @endcode
  The reason is that the server will optimize for cases of only a single
  record. If, in a table scan, you don't know the number of records, it
  will probably be better to set records to two so you can return as many
  records as you need. Along with records, a few more variables you may wish
  to set are:
    records
    deleted
    data_file_length
    index_file_length
    delete_length
    check_time
  Take a look at the public variables in handler.h for more information.

  Called in filesort.cc, ha_heap.cc, item_sum.cc, opt_sum.cc, sql_delete.cc,
  sql_delete.cc, sql_derived.cc, sql_select.cc, sql_select.cc, sql_select.cc,
  sql_select.cc, sql_select.cc, sql_show.cc, sql_show.cc, sql_show.cc,
  sql_show.cc, sql_table.cc, sql_union.cc, and sql_update.cc.

    @see
  filesort.cc, ha_heap.cc, item_sum.cc, opt_sum.cc, sql_delete.cc,
  sql_select.cc, sql_table.cc, sql_union.cc and sql_update.cc
*/
int ha_revision::info(uint flag)
{
  DBUG_ENTER("ha_revision::info");
  int r= uhandler->get_uhandler()->info(flag);
  if (flag & HA_STATUS_VARIABLE)
  {
    stats.records= uhandler->get_uhandler()->stats.records;
    stats.deleted= uhandler->get_uhandler()->stats.deleted;
    stats.data_file_length= uhandler->get_uhandler()->stats.data_file_length;
    stats.index_file_length= uhandler->get_uhandler()->stats.index_file_length;
    stats.delete_length= uhandler->get_uhandler()->stats.delete_length;
    stats.check_time= uhandler->get_uhandler()->stats.check_time;
    stats.mean_rec_length= uhandler->get_uhandler()->stats.mean_rec_length;
  }
  if (flag & HA_STATUS_CONST)
  {
    stats.max_data_file_length=
      uhandler->get_uhandler()->stats.max_data_file_length;
    stats.max_index_file_length=
      uhandler->get_uhandler()->stats.max_index_file_length;
    stats.create_time= uhandler->get_uhandler()->stats.create_time;
    ref_length= uhandler->get_uhandler()->ref_length;
    stats.block_size= uhandler->get_uhandler()->stats.block_size;
  }
  if (flag & HA_STATUS_TIME)
    stats.update_time = uhandler->get_uhandler()->stats.update_time;
  if (flag & HA_STATUS_AUTO)
    stats.auto_increment_value=
      uhandler->get_uhandler()->stats.auto_increment_value;
  DBUG_RETURN(r);
}


/** @brief
  extra() is called whenever the server wishes to send a hint to
  the storage engine. The myisam engine implements the most hints.
  ha_innodb.cc has the most exhaustive list of these hints.

    @see
  ha_innodb.cc
*/
int ha_revision::extra(enum ha_extra_function operation)
{
  DBUG_ENTER("ha_revision::extra");
  DBUG_RETURN(uhandler->get_uhandler()->extra(operation));
}


void ha_revision::get_auto_increment(ulonglong offset, ulonglong increment,
                                     ulonglong nb_desired_values,
                                     ulonglong *first_value,
                                     ulonglong *nb_reserved_values)
{
  uhandler->get_uhandler()->get_auto_increment(offset, increment,
                                               nb_desired_values,
                                               first_value,
                                               nb_reserved_values);
}


int ha_revision::reset()
{
  return uhandler->reset();
}

/*
  Reset the auto-increment counter to the given value, i.e. the next row
  inserted will get the given value. This is called e.g. after TRUNCATE
  is emulated by doing a 'DELETE FROM t'. HA_ERR_WRONG_COMMAND is
  returned by storage engines that don't support this operation.
*/
int ha_revision::reset_auto_increment(ulonglong value)
{
  return uhandler->get_uhandler()->ha_reset_auto_increment(value);
}


void ha_revision::update_create_info(HA_CREATE_INFO *create_info)
{
  uhandler->get_uhandler()->update_create_info(create_info);
}


void ha_revision::release_auto_increment()
{
  uhandler->get_uhandler()->ha_release_auto_increment();
}


/** @brief
  Used to delete all rows in a table, including cases of truncate and cases
  where the optimizer realizes that all rows will be removed as a result of
  an SQL statement.

    @details
  Called from item_sum.cc by Item_func_group_concat::clear(),
  Item_sum_count_distinct::clear(), and Item_func_group_concat::clear().
  Called from sql_delete.cc by mysql_delete().
  Called from sql_select.cc by JOIN::reinit().
  Called from sql_union.cc by st_select_lex_unit::exec().

    @see
  Item_func_group_concat::clear(), Item_sum_count_distinct::clear() and
  Item_func_group_concat::clear() in item_sum.cc;
  mysql_delete() in sql_delete.cc;
  JOIN::reinit() in sql_select.cc and
  st_select_lex_unit::exec() in sql_union.cc.
*/
int ha_revision::delete_all_rows()
{
  DBUG_ENTER("ha_revision::delete_all_rows");
  DBUG_RETURN(uhandler->get_uhandler()->ha_delete_all_rows());
}


uint ha_revision::lock_count(void) const
{
  return uhandler->lock_count();
}


/** @brief
  This create a lock on the table. If you are implementing a storage engine
  that can handle transacations look at ha_berkely.cc to see how you will
  want to go about doing this. Otherwise you should consider calling flock()
  here. Hint: Read the section "locking functions for mysql" in lock.cc to
  understand this.

    @details
  Called from lock.cc by lock_external() and unlock_external(). Also called
  from sql_table.cc by copy_data_between_tables().

    @see
  lock.cc by lock_external() and unlock_external() in lock.cc;
  the section "locking functions for mysql" in lock.cc;
  copy_data_between_tables() in sql_table.cc.
*/
int ha_revision::external_lock(THD *thd, int lock_type)
{
  DBUG_ENTER("ha_revision::external_lock");
  DBUG_RETURN(uhandler->external_lock(thd, lock_type));
}


/** @brief
  The idea with handler::store_lock() is: The statement decides which locks
  should be needed for the table. For updates/deletes/inserts we get WRITE
  locks, for SELECT... we get read locks.

    @details
  Before adding the lock into the table lock handler (see thr_lock.c),
  mysqld calls store lock with the requested locks. Store lock can now
  modify a write lock to a read lock (or some other lock), ignore the
  lock (if we don't want to use MySQL table locks at all), or add locks
  for many tables (like we do when we are using a MERGE handler).

  Berkeley DB, for exmaple, changes all WRITE locks to TL_WRITE_ALLOW_WRITE
  (which signals that we are doing WRITES, but are still allowing other
  readers and writers).

  When releasing locks, store_lock() is also called. In this case one
  usually doesn't have to do anything.

  In some exceptional cases MySQL may send a request for a TL_IGNORE;
  This means that we are requesting the same lock as last time and this
  should also be ignored. (This may happen when someone does a flush
  table when we have opened a part of the tables, in which case mysqld
  closes and reopens the tables and tries to get the same locks at last
  time). In the future we will probably try to remove this.

  Called from lock.cc by get_lock_data().

    @see
  get_lock_data() in lock.cc
*/
THR_LOCK_DATA **ha_revision::store_lock(THD *thd,
                                       THR_LOCK_DATA **to,
                                       enum thr_lock_type lock_type)
{
  return uhandler->store_lock(thd, to, lock_type);
}


void get_comment_from_rev_file(const char *name,
                               LEX_STRING& nm )
{
  char filename[FN_REFLEN];
  fn_format(filename, name, "", ha_revision_exts[0],
            MY_UNPACK_FILENAME | MY_APPEND_EXT);
  File file;
  if ((file= my_open(filename, O_RDONLY, MYF(MY_WME))) < 0)
    return;

  if ((nm.length= my_read(file, reinterpret_cast<uchar*>(nm.str),
                          nm.length, MYF(MY_WME))) == 0)
    return;

  nm.str[nm.length]= 0;
  my_close(file, MYF(MY_WME));
}


handlerton *ha_revision::get_handlerton_from_rev_file(const char *name,
                                                      LEX_STRING& nm )
{

  get_comment_from_rev_file(name, nm);
  if (nm.length == 0)
    return NULL;

  THD *thd= current_thd;
  return plugin_data(ha_resolve_by_name(thd, &nm), handlerton *);
}


/** @brief
  Used to delete a table. By the time delete_table() has been called all
  opened references to this table will have been closed (and your globally
  shared references released). The variable name will just be the name of
  the table. You will need to remove any files you have created at this point.

    @details
  If you do not implement this, the default delete_table() is called from
  handler.cc and it will delete all files with the file extensions returned
  by bas_ext().

  Called from handler.cc by delete_table and ha_create_table(). Only used
  during create if the table_flag HA_DROP_BEFORE_CREATE was specified for
  the storage engine.

    @see
  delete_table and ha_create_table() in handler.cc
*/
int ha_revision::delete_table(const char *name)
{
  DBUG_ENTER("ha_revision::delete_table");

  THD *thd = current_thd;

  TABLE_SHARE table;
  bzero(static_cast<void *>(&table), sizeof(table));
  const uint size = 32;
  char type[size];
  LEX_STRING nm = { type, size };

  get_comment_from_rev_file(name, nm);
  table.comment= nm;
  uhandler= instantiate_revision_handler(&table,
                                         thd->mem_root);
  uhandler->delete_table(name, &table);
  DBUG_RETURN(handler::delete_table(name));
}


/** @brief
  Renames a table from one name to another via an alter table call.

    @details
  If you do not implement this, the default rename_table() is called from
  handler.cc and it will delete all files with the file extensions returned
  by bas_ext().

  Called from sql_table.cc by mysql_rename_table().

    @see
  mysql_rename_table() in sql_table.cc
*/
int ha_revision::rename_table(const char * from, const char * to)
{
  DBUG_ENTER("ha_revision::rename_table ");

  THD *thd = current_thd;
  const uint size = 32;
  char type[size];
  LEX_STRING nm = { type, size };
  handlerton *underlying_from= get_handlerton_from_rev_file(from, nm);

  if (underlying_from == NULL)
  {
    char warn_buff[MYSQL_ERRMSG_SIZE];
    my_snprintf(warn_buff, sizeof(warn_buff), ER(ER_UNKNOWN_STORAGE_ENGINE),
                nm.str);
    push_warning(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                 ER_UNKNOWN_STORAGE_ENGINE, warn_buff);
    DBUG_RETURN(ER_UNKNOWN_STORAGE_ENGINE);
  }
  else
  {
    TABLE_SHARE table;
    bzero(static_cast<void *>(&table), sizeof(table));

    handler *uhandler= underlying_from->create(underlying_from,
                                               &table,
                                               thd->mem_root);
    uhandler->init();
    int r= uhandler->ha_rename_table(from, to);
    delete uhandler;
    if (r != 0)
      DBUG_RETURN(r);

    /* Rename .REV file as well.
    */
    char old_name[FN_REFLEN], new_name[FN_REFLEN];
    fn_format(old_name, from, "", ha_revision_exts[0],
              MY_UNPACK_FILENAME | MY_APPEND_EXT);
    fn_format(new_name, to, "", ha_revision_exts[0],
              MY_UNPACK_FILENAME | MY_APPEND_EXT);
    if (my_rename_with_symlink(old_name, new_name, MYF(MY_WME)))
      DBUG_RETURN(my_errno);

    DBUG_RETURN(0);
  }
}


/** @brief
  Given a starting key and an ending key, estimate the number of rows that
  will exist between the two keys.

    @details
  end_key may be empty, in which case determine if start_key matches any rows.

  Called from opt_range.cc by check_quick_keys().

    @see
  check_quick_keys() in opt_range.cc
*/
ha_rows ha_revision::records_in_range(uint inx, key_range *min_key,
                                      key_range *max_key)
{
  DBUG_ENTER("ha_revision::records_in_range");
  DBUG_RETURN(uhandler->get_uhandler()->records_in_range(inx,
                                                         min_key, max_key));
}


/** @brief
  create() is called to create a database. The variable name will have the name
  of the table.

    @details
  When create() is called you do not need to worry about opening the table.
  Also, the .frm file will have already been created so adjusting create_info
  is not necessary. You can overwrite the .frm file at this point if you wish
  to change the table definition, but there are no methods currently provided
  for doing so.

  Called from handle.cc by ha_create_table().

    @see
  ha_create_table() in handle.cc
*/
int ha_revision::create(const char *name, TABLE *table_arg,
                        HA_CREATE_INFO *create_info)
{
  DBUG_ENTER("ha_revision::create");

  /* Instantiate the right revision handler depending on the comment field.
  */
  uhandler= instantiate_revision_handler(table_arg->s, mem_root);

  int r= uhandler->create(name, table_arg, create_info);
  if (r != 0)
    DBUG_RETURN(r);

  /* Write storage engine to a file (for delete table).
  */
  char filename[FN_REFLEN];
  fn_format(filename, name, "", ha_revision_exts[0],
            MY_UNPACK_FILENAME | MY_APPEND_EXT);
  File file;
  if ((file= my_create(filename, 0, O_RDWR | O_TRUNC, MYF(MY_WME))) < 0)
  {
    DBUG_RETURN(my_errno);
  }

  if (table_arg->s->comment.length == 0)
  {
    const char *def=
      ha_resolve_storage_engine_name(ha_default_handlerton(current_thd));
    uint len= strlen(def);
    if (my_write(file, reinterpret_cast<const uchar*>(def),
                 len, MYF(MY_WME)) < len)
    {
      my_close(file, MYF(MY_WME));
      DBUG_RETURN(-1);
    }
  }
  else
  {

    if (my_write(file,
                 reinterpret_cast<const uchar*>(table_arg->s->comment.str),
                 table_arg->s->comment.length,
                 MYF(MY_WME)) < table_arg->s->comment.length)
    {
      my_close(file, MYF(MY_WME));
      DBUG_RETURN(-1);
    }
  }
  my_close(file, MYF(MY_WME));

  DBUG_RETURN(r);
}


static int check_select_mode(MYSQL_THD,
                             struct st_mysql_sys_var *,
                             void *save, struct st_mysql_value *value)
{
  char *buffer= 0;
  int length= 0;
  const char *v= value->val_str(value, buffer, &length);
  *static_cast<void **>(save)= my_malloc(sizeof(char) * (length + 1),
                                         MYF(MY_WME));
  if (*static_cast<char **>(save))
    memcpy(*static_cast<char **>(save), v, length + 1);
  return 0;
}


static struct st_mysql_sys_var* revision_system_variables[]= {
  MYSQL_SYSVAR(select_mode),
  NULL
};


struct st_mysql_storage_engine revision_storage_engine=
{ MYSQL_HANDLERTON_INTERFACE_VERSION };


mysql_declare_plugin(revision)
{
  MYSQL_STORAGE_ENGINE_PLUGIN,
  &revision_storage_engine,
  "REVISION",
  "Guillaume d'Humieres, DDEngine.org",
  "Revision storage engine",
  PLUGIN_LICENSE_GPL,
  revision_init_func, /* Plugin Init */
  revision_done_func, /* Plugin Deinit */
  0x0001 /* 0.1 */,
  NULL,                       /* status variables                */
  revision_system_variables,  /* system variables                */
  NULL                        /* config options                  */
}
mysql_declare_plugin_end;
