/* Copyright (C) 2007 DDengine.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef _revision_handler_h
#define _revision_handler_h

const char select_current[] = "current";
const char select_history[] = "history";
const char select_deleted[] = "deleted";

/** @brief
  Revision handler interface.
*/
class revision_handler
{

public:
  /** Which row(s) SELECT statements should return.
      Default is current.
  */
  enum select_mode_t
  {
    none = 0, ///< No specific mode (default for update/delete statements)
    current,  ///< Only the current row (select statements)
    history,  ///< All rows
    deleted,  ///< Only deleted rows
    timestamp ///< Data set at a specific timestamp
  };

  virtual ~revision_handler() { }

  /** Used for opening tables. The name will be the name of the file.
  */
  virtual int open(TABLE *, const char *name,
                   int mode, uint test_if_locked) = 0;

  /** Closes a table.
  */
  virtual int close() = 0;

  /** Reset state of file to after 'open'
      This function is called after every statement for all tables used
      by that statement.
  */
  virtual int reset() = 0;

  /** Creates the table.
  */
  virtual int create(const char *name, TABLE *table_arg,
                     HA_CREATE_INFO *create_info) = 0;

  /** Inserts a row.
  */
  virtual int write_row(uchar * buf) = 0;

  /** Updates the contents of an existing row.
  */
  virtual int update_row(const uchar * old_data, uchar * new_data) = 0;

  /** Deletes a row.
  */
  virtual int delete_row(const uchar * buf) = 0;

  /** Handler-specific error messages.
  */
  virtual bool get_error_message(int error, String *buf) = 0;

  /* Returns the underlying handler.
  */
  virtual handler *get_uhandler() = 0;

  /** Index read method.
  */
  virtual int index_read(uchar * buf, const uchar * key,
                         uint key_len, enum ha_rkey_function find_flag) = 0;

  /** Used to read forward through the index.
  */
  virtual int index_next(uchar * buf) = 0;

  /** Table scan method.
  */
  virtual int rnd_next(uchar *buf) = 0;

  /** Called in test_quick_select to determine if indexes should be used.
  */
  virtual double read_time(uint index, uint ranges, ha_rows rows) = 0;

  /** Called in test_quick_select to determine if indexes should be used.
  */
  virtual double scan_time() = 0;

  virtual uint lock_count(void) const = 0;

  virtual int external_lock(THD *thd, int lock_type) = 0;

  virtual THR_LOCK_DATA **store_lock(THD *thd, THR_LOCK_DATA **to,
                                     enum thr_lock_type lock_type) = 0;
  /** Sets the select mode for subsequent queries.
  */
  virtual void set_select_mode(select_mode_t) = 0;

  /** Sets the timestamp to be used when data state at a specific time
      is to be returned.
  */
  virtual void set_timestamp(const MYSQL_TIME &) = 0;

  /** Delete tables.
  */
  virtual void delete_table(const char *, TABLE_SHARE *) = 0;
};

#endif
